@@ -3,26 +3,28 @@ from __future__ import annotations
 from dataclasses import dataclass
 from itertools import zip_longest
 from math import floor, sqrt
-from typing import Any, Literal, Sequence
+from typing import Any, Literal, Sequence, Tuple
 
 from vsexprtools import ExprList, ExprOp, ExprToken, aka_expr_available, norm_expr
 from vsrgtools.util import wmean_matrix
 from vstools import (
     ConvMode, CustomIndexError, FuncExceptT, PlanesT, StrList, check_variable, copy_signature, core, fallback,
-    inject_self, interleave_arr, iterate, vs
+    inject_self, interleave_arr, iterate, scale_value, to_arr, vs
 )
 
 from .types import Coordinates, MorphoFunc, XxpandMode
 
 __all__ = [
-    'Morpho',
+    'Morpho', 'CoordsT',
     'grow_mask'
 ]
 
+CoordsT = int | Tuple[int, ConvMode] | Sequence[int]
+
 
 def _minmax_method(  # type: ignore
-    self: Morpho, src: vs.VideoNode, thr: int | float | None = None,
-    coords: int | tuple[int, ConvMode] | Sequence[int] | None = [1] * 8,
+    self: Morpho, src: vs.VideoNode, thr: float | None = None,
+    coords: CoordsT | None = [1] * 8,
     iterations: int = 1, multiply: float | None = None, planes: PlanesT = None,
     *, func: FuncExceptT | None = None, **kwargs: Any
 ) -> vs.VideoNode:
@@ -30,8 +32,8 @@ def _minmax_method(  # type: ignore
 
 
 def _morpho_method(  # type: ignore
-    self: Morpho, src: vs.VideoNode, radius: int = 1, planes: PlanesT = None, thr: int | float | None = None,
-    coords: int | tuple[int, ConvMode] | Sequence[int] = 5, multiply: float | None = None,
+    self: Morpho, src: vs.VideoNode, radius: int = 1, planes: PlanesT = None, thr: float | None = None,
+    coords: CoordsT = 5, multiply: float | None = None,
     *, func: FuncExceptT | None = None, **kwargs: Any
 ) -> vs.VideoNode:
     ...
@@ -39,7 +41,7 @@ def _morpho_method(  # type: ignore
 
 def _morpho_method2(  # type: ignore
     self: Morpho, clip: vs.VideoNode, sw: int, sh: int | None = None, mode: XxpandMode = XxpandMode.RECTANGLE,
-    thr: int | None = None, planes: PlanesT = None, *, func: FuncExceptT | None = None, **kwargs: Any
+    thr: float | None = None, planes: PlanesT = None, *, func: FuncExceptT | None = None, **kwargs: Any
 ) -> vs.VideoNode:
     ...
 
@@ -54,8 +56,7 @@ class Morpho:
         self._fast = fallback(self.fast, aka_expr_available) and aka_expr_available
 
     def _check_params(
-        self, radius: int, thr: int | float | None, coords: int | tuple[int, ConvMode] | Sequence[int],
-        planes: PlanesT, func: FuncExceptT
+        self, radius: int, thr: float | None, coords: CoordsT, planes: PlanesT, func: FuncExceptT
     ) -> tuple[FuncExceptT, PlanesT]:
         if radius < 1:
             raise CustomIndexError('radius has to be greater than 0!', func, radius)
@@ -80,9 +81,8 @@ class Morpho:
 
     @classmethod
     def _morpho_xx_imum(
-        cls, thr: int | float | None, op: Literal[ExprOp.MIN, ExprOp.MAX],
-        coords: int | tuple[int, ConvMode] | Sequence[int], multiply: float | None = None,
-        clamp: bool = False
+        cls, src: vs.VideoNode, thr: float | None, op: Literal[ExprOp.MIN, ExprOp.MAX],
+        coords: CoordsT, multiply: float | None = None, clamp: bool = False
     ) -> StrList:
         exclude = list[tuple[int, int]]()
 
@@ -112,7 +112,7 @@ class Morpho:
         matrix = ExprList(interleave_arr(matrix, op * matrix.mlength, 2))
 
         if thr is not None:
-            matrix.append('x', thr, ExprOp.SUB, ExprOp.MAX)
+            matrix.append('x', scale_value(thr, 32, src), ExprOp.SUB, ExprOp.MAX)
 
         if multiply is not None:
             matrix.append(multiply, ExprOp.MUL)
@@ -123,16 +123,15 @@ class Morpho:
         return matrix
 
     def _mm_func(
-        self, src: vs.VideoNode, radius: int = 1, planes: PlanesT = None, thr: int | float | None = None,
-        coords: int | tuple[int, ConvMode] | Sequence[int] = 5, multiply: float | None = None,
-        *, func: FuncExceptT, mm_func: MorphoFunc, op: Literal[ExprOp.MIN, ExprOp.MAX],
-        **kwargs: Any
+        self, src: vs.VideoNode, radius: int = 1, planes: PlanesT = None, thr: float | None = None,
+        coords: CoordsT = 5, multiply: float | None = None, *, func: FuncExceptT,
+        mm_func: MorphoFunc, op: Literal[ExprOp.MIN, ExprOp.MAX], **kwargs: Any
     ) -> vs.VideoNode:
         func, planes = self._check_params(radius, thr, coords, planes, func)
 
         if self._fast:
             mm_func = norm_expr  # type: ignore[assignment]
-            kwargs.update(expr=self._morpho_xx_imum(thr, op, coords, multiply))
+            kwargs.update(expr=self._morpho_xx_imum(src, thr, op, coords, multiply))
         elif isinstance(coords, (int, tuple)):
             if isinstance(coords, tuple):
                 if coords[1] is not ConvMode.SQUARE:
@@ -147,23 +146,27 @@ class Morpho:
                     'with fast=False or no akarin plugin, you must have coords=3!', func, coords
                 )
 
-            kwargs.update(coords=[1] * 8)
+            kwargs.update(coordinates=[1] * 8)
 
-        if not self._fast and multiply is not None:
-            orig_mm_func = mm_func
+        if not self._fast:
+            if thr is not None:
+                kwargs.update(threshold=scale_value(thr, 32, src))
 
-            @copy_signature(mm_func)
-            def _mm_func(*args: Any, **kwargs: Any) -> Any:
-                return orig_mm_func(*args, **kwargs).std.Expr(f'x {multiply} *')
+            if multiply is not None:
+                orig_mm_func = mm_func
 
-            mm_func = _mm_func
+                @copy_signature(mm_func)
+                def _mm_func(*args: Any, **kwargs: Any) -> Any:
+                    return orig_mm_func(*args, **kwargs).std.Expr(f'x {multiply} *')
+
+                mm_func = _mm_func
 
         return iterate(src, mm_func, radius, planes=planes, **kwargs)
 
     @inject_self
     def xxpand_transform(
         self, clip: vs.VideoNode, op: Literal[ExprOp.MIN, ExprOp.MAX], sw: int, sh: int | None = None,
-        mode: XxpandMode = XxpandMode.RECTANGLE, thr: int | None = None,
+        mode: XxpandMode = XxpandMode.RECTANGLE, thr: float | None = None,
         planes: PlanesT = None, *, func: FuncExceptT | None = None
     ) -> vs.VideoNode:
         func, planes = self._check_params(1, thr, 3, planes, func or self.xxpand_transform)
@@ -190,7 +193,7 @@ class Morpho:
         return clip
 
     def _xxflate(
-        self: Morpho, inflate: bool, src: vs.VideoNode, radius: int, planes: PlanesT, thr: int | float | None,
+        self: Morpho, inflate: bool, src: vs.VideoNode, radius: int, planes: PlanesT, thr: float | None,
         multiply: float | None, *, func: FuncExceptT
     ) -> vs.VideoNode:
         assert check_variable(src, func)
@@ -208,6 +211,7 @@ class Morpho:
         expr.append('x', ExprOp.MAX if inflate else ExprOp.MIN)
 
         if thr is not None:
+            thr = scale_value(thr, 32, src)
             limit = ['x', thr, ExprOp.ADD] if inflate else ['x', thr, ExprOp.SUB, ExprToken.RangeMin, ExprOp.MAX]
 
             expr.append(limit, ExprOp.MIN if inflate else ExprOp.MAX)
@@ -220,8 +224,7 @@ class Morpho:
     @inject_self
     @copy_signature(_minmax_method)
     def maximum(
-        self, src: vs.VideoNode, thr: int | float | None = None,
-        coords: int | tuple[int, ConvMode] | Sequence[int] | None = None,
+        self, src: vs.VideoNode, thr: float | None = None, coords: CoordsT | None = None,
         iterations: int = 1, multiply: float | None = None, planes: PlanesT = None,
         *, func: FuncExceptT | None = None, **kwargs: Any
     ) -> vs.VideoNode:
@@ -230,8 +233,7 @@ class Morpho:
     @inject_self
     @copy_signature(_minmax_method)
     def minimum(
-        self, src: vs.VideoNode, thr: int | float | None = None,
-        coords: int | tuple[int, ConvMode] | Sequence[int] | None = None,
+        self, src: vs.VideoNode, thr: float | None = None, coords: CoordsT | None = None,
         iterations: int = 1, multiply: float | None = None, planes: PlanesT = None,
         *, func: FuncExceptT | None = None, **kwargs: Any
     ) -> vs.VideoNode:
@@ -239,7 +241,7 @@ class Morpho:
 
     @inject_self
     def inflate(
-        self: Morpho, src: vs.VideoNode, radius: int = 1, planes: PlanesT = None, thr: int | float | None = None,
+        self: Morpho, src: vs.VideoNode, radius: int = 1, planes: PlanesT = None, thr: float | None = None,
         iterations: int = 1, multiply: float | None = None, *, func: FuncExceptT | None = None
     ) -> vs.VideoNode:
         for _ in range(iterations):
@@ -248,7 +250,7 @@ class Morpho:
 
     @inject_self
     def deflate(
-        self: Morpho, src: vs.VideoNode, radius: int = 1, planes: PlanesT = None, thr: int | float | None = None,
+        self: Morpho, src: vs.VideoNode, radius: int = 1, planes: PlanesT = None, thr: float | None = None,
         iterations: int = 1, multiply: float | None = None, *, func: FuncExceptT | None = None
     ) -> vs.VideoNode:
         for _ in range(iterations):
@@ -297,17 +299,16 @@ class Morpho:
 
     @inject_self
     def gradient(
-        self, src: vs.VideoNode, radius: int = 1, planes: PlanesT = None, thr: int | float | None = None,
-        coords: int | tuple[int, ConvMode] | Sequence[int] = 5, multiply: float | None = None,
-        *, func: FuncExceptT | None = None, **kwargs: Any
+        self, src: vs.VideoNode, radius: int = 1, planes: PlanesT = None, thr: float | None = None,
+        coords: CoordsT = 5, multiply: float | None = None, *, func: FuncExceptT | None = None, **kwargs: Any
     ) -> vs.VideoNode:
         func, planes = self._check_params(radius, thr, coords, planes, func or self.gradient)
 
         if radius == 1 and self._fast:
             return norm_expr(
                 src, '{dilated} {eroded} - {multiply}', planes,
-                dilated=self._morpho_xx_imum(thr, ExprOp.MAX, coords, None, True),
-                eroded=self._morpho_xx_imum(thr, ExprOp.MIN, coords, None, True),
+                dilated=self._morpho_xx_imum(src, thr, ExprOp.MAX, coords, None, True),
+                eroded=self._morpho_xx_imum(src, thr, ExprOp.MIN, coords, None, True),
                 multiply='' if multiply is None else f'{multiply} *'
             )
 
@@ -332,16 +333,15 @@ class Morpho:
 
     @inject_self
     def outer_hat(
-        self, src: vs.VideoNode, radius: int = 1, planes: PlanesT = None, thr: int | float | None = None,
-        coords: int | tuple[int, ConvMode] | Sequence[int] = 5, multiply: float | None = None,
-        *, func: FuncExceptT | None = None, **kwargs: Any
+        self, src: vs.VideoNode, radius: int = 1, planes: PlanesT = None, thr: float | None = None,
+        coords: CoordsT = 5, multiply: float | None = None, *, func: FuncExceptT | None = None, **kwargs: Any
     ) -> vs.VideoNode:
         func, planes = self._check_params(radius, thr, coords, planes, func or self.outer_hat)
 
         if radius == 1 and self._fast:
             return norm_expr(
                 src, '{dilated} {multiply} x -', planes,
-                dilated=self._morpho_xx_imum(thr, ExprOp.MAX, coords, None, True),
+                dilated=self._morpho_xx_imum(src, thr, ExprOp.MAX, coords, None, True),
                 multiply='' if multiply is None else f'{multiply} *'
             )
 
@@ -351,16 +351,15 @@ class Morpho:
 
     @inject_self
     def inner_hat(
-        self, src: vs.VideoNode, radius: int = 1, planes: PlanesT = None, thr: int | float | None = None,
-        coords: int | tuple[int, ConvMode] | Sequence[int] = 5, multiply: float | None = None,
-        *, func: FuncExceptT | None = None, **kwargs: Any
+        self, src: vs.VideoNode, radius: int = 1, planes: PlanesT = None, thr: float | None = None,
+        coords: CoordsT = 5, multiply: float | None = None, *, func: FuncExceptT | None = None, **kwargs: Any
     ) -> vs.VideoNode:
         func, planes = self._check_params(radius, thr, coords, planes, func or self.inner_hat)
 
         if radius == 1 and self._fast:
             return norm_expr(
                 src, '{eroded} {multiply} x -', planes,
-                eroded=self._morpho_xx_imum(thr, ExprOp.MIN, coords),
+                eroded=self._morpho_xx_imum(src, thr, ExprOp.MIN, coords),
                 multiply='' if multiply is None else f'{multiply} *'
             )
 
@@ -368,19 +367,34 @@ class Morpho:
 
         return norm_expr([eroded, src], 'x y -', planes)
 
+    @inject_self
+    def binarize(
+        self, src: vs.VideoNode, midthr: float | list[float] | None = None,
+        lowval: float | list[float] | None = None, highval: float | list[float] | None = None,
+        planes: PlanesT = None
+    ) -> vs.VideoNode:
+        midthr, lowval, highval = (
+            thr and list(
+                scale_value(t, 32, src, chroma=i != 0)
+                for i, t in enumerate(to_arr(thr))
+            ) for thr in (midthr, lowval, highval)
+        )
+
+        return src.std.Binarize(midthr, lowval, highval, planes)
+
 
 def grow_mask(
     mask: vs.VideoNode, radius: int = 1, multiply: float = 1.0,
-    planes: PlanesT = None, coords: int | tuple[int, ConvMode] | Sequence[int] = 5,
-    thr: int | float | None = None, *, func: FuncExceptT | None = None, **kwargs: Any
+    planes: PlanesT = None, coords: CoordsT = 5, thr: float | None = None,
+    *, func: FuncExceptT | None = None, **kwargs: Any
 ) -> vs.VideoNode:
     func = func or grow_mask
 
     assert check_variable(mask, func)
 
-    morpho = Morpho(planes)
+    morpho = Morpho(planes, func)
 
-    kwargs.update(thr=thr, coords=coords, func=func)
+    kwargs.update(thr=thr, coords=coords)
 
     closed = morpho.closing(mask, **kwargs)
     dilated = morpho.dilation(closed, **kwargs)
